/*
 * Antoun Kanawati
 * November 27, 1989
 *
 * Another DOS hacking program.
 *
 *                    A program to trace DOS's MCB chain.
 *
 * A Memory Control Block (MCB) is an undocumented DOS data structure built as 
 * follows:
 *    size      = 16 bytes.
 *    offset    contents
 *      0       byte   'M' ==> middle of chain.
 *                     'Z' ==> end of chain.
 *      1       word    0  ==> Free.
 *                   != 0  ==> Segment address of owner (PSP address).
 *      3       word       ==> Size of block.
 *      4       11 bytes   ==> reserved or unused.
 *
 * For each block that DOS allocates thru function 0x48, one of these blocks
 * is created, an it is just before the block whose address is returned.
 * (This is true even for blocks of size 0.)
 *
 * Furthermore, all MCB's are aligned on paragraph boundaries.
 *
 * The program name is printed by looking in the owner's PSP, and scanning the
 * owner's environment block (if it is still attached to the program).
 *
 * For details on how the environment is located, look at ADVANCED MS-DOS, or
 * the MS-DOS encyclopedia.
 *
 * The MCB structure is sketched in a few non-official references.  Official 
 * MS-DOS references (such as the two above) barely mention it.
 * None of the references I have seen mention how to find the starting MCB, 
 * so I had to write an educated guessing routine to do it.
 *
 * A comment on the use of the program:
 *  The casual user and the high level programmer will find this stuff of no 
 *  use.  This is intended for those who like to know what happens for the 
 *  sake of knowing what happens (i.e.: for the fun of it).  There are 
 *  programs of a higher quality output and usefulness available to list all
 *  your memory resident programs, the interrupts they use, etc....
 *  These other niceties are trivial problems once we have traced the chain:
 *  sort the blocks by segment address, and scan the vector table checking the
 *  segment addresses against the owners of MCB's.
 *
 * Disclaimer:
 *  You may use this program at your own risk.  The author is not responsible 
 *  in any manner for damages incurred by the use of the program, nor for
 *  benefits gained from the source code.  That is: feel free to copy, modify, 
 *  and distribute the source code, but do not use it blindly -- This is 
 *  intended for competent programmers with substantial DOS knowledge.  And 
 *  please keep the source code of this program with to its binaries -- The
 *  executable is quite worthless without the source code.
 *
 */
#include <stdio.h>

#if sizeof(char *) != 4
  /*
   * The program must operate with 32-bit pointers.
   *
   */
main()
{
  printf("This program must be compiled in a large data pointer model.\n");
  exit(0);
}
#else

   /* Going from segment addresses to normalized pointer values. */
#define seg_to_charp(x) ((char *) (((long) (x)) << 16))

   /* MCB access protocols: */
#define MCB_owner(ptr)       (*((unsigned *) (((char *) (ptr))+1)))
#define MCB_size(ptr)        (*((unsigned *) (((char *) (ptr))+3)))

#define BIOS_END             0x80

unsigned start;

/*
 * Environment scanner:
 *
 *   Given the PSP of the owner checks:
 *      1- PSP starts with a INT 20H instruction.
 *      2- Retrieves the environment block address, and checks that it is
 *         attached to the program.
 *   Then, the evironment is scanned for the program name, which is printed.
 *
 */
void do_env(unsigned psp)
{
  char     *envp;
  char     *psp_ptr;
  unsigned envseg;
  unsigned envsize;


    /* The first MCB has invalid owner/environment inforamtion. */
  if (psp == start)
    {
      printf("* Start MCB.\n");
      return;
    }

  psp_ptr = seg_to_charp(psp);

    /* If the PSP does not start with INT 20H, return. */
  if ((*((unsigned *) psp_ptr)) != 0x20CD) /* INT 20H */
    {
      printf("* PSP does not start with INT 20H\n");
      return;
    }

    /* Look at the MCB of environment block : */
  envseg = *((unsigned *) (psp_ptr + 0x2C));
  envp   = seg_to_charp(envseg-1);


    /* The MCB must be an M-type, owned by PSP that was used to reach it, */
    /* and must have a non-zero size.                                     */
    /* If any of these conditions fails, we leave.                        */
  if ( *envp != 'M' || MCB_owner(envp) != psp || MCB_size(envp) == 0)
    {
      if (*envp != 'M') printf("* Env MCB reused ");
      if (MCB_owner(envp) != psp) printf("* Env owner changed ");
      if (MCB_size(envp) == 0) printf("* Env size == 0");
      printf("\n");
      return;
    }


  envsize = MCB_size(envp) << 4; /* paragraphs --> bytes. */

    /* Now, we skip the environment strings, which leaves us at the program */
    /* name.                                                                */
  envp = seg_to_charp(envseg);

  while (*envp && envsize)
    while (*envp++ && envsize--)
      ;

  envp    += 3;
  envsize -= 4; /* last envsize-- is not executed in normal circumstances. */

    /* If the whole environment was skipped, we print nothing. */
  if (envsize > 0)
    printf("%s\n", envp);
  else
    printf("* No name found\n");

}



/*
 * Prints contents of MCB:
 *   1- Check for valid format.
 *   2- Pick up the data on size an owner.
 *   3- Print.
 *   4- Try to find an owner's name.
 *
 */
void dump_mcb(unsigned seg_ref)
{
  char      *p;
  unsigned  owner;
  unsigned  size;

  p = seg_to_charp(seg_ref);

  if (*p != 'M' && *p != 'Z')
    {
      printf("Bad MCB, first byte == %02X\n", (unsigned char) (*p));
      return;
    }

    /* The first MCB has invalid owner/environment inforamtion. */
  if (seg_ref == start)
    {
       printf("%04X  %02X                  * First MCB\n",
               seg_ref,
               (unsigned char) (*p));
       return;
    }


  owner = MCB_owner(p);
  size  = MCB_size(p);

  printf("%04X  %02X  %04X  %04X %04X ",
          seg_ref,
          (unsigned char) (*p),
          owner,
          *((unsigned *) (seg_to_charp(owner) + 0x16)),
          size);

  if (*p == 'Z')
    printf("* End of chain.\n");
  else if (owner == 0)
    printf("* Free fragment.\n");
  else
    do_env(owner);
}


/*
 * Forward tracing in the allocation chain.
 *
 * Note that we assume that we have a proper MCB address at start.
 *
 */
void trace_chain(unsigned mcb_seg)
{
  unsigned seg;
  char *p;

  seg = mcb_seg;

  printf("Addr LBL  OWNR PARNT SIZE\n");
  printf("---- ---- ---- ----- ----\n");


  for(;;)
  {
    dump_mcb(seg);          /* Print the current MCB. */

    p = seg_to_charp(seg);

    if (*p == 'Z') break;   /* End of chain ? */

    if (*p != 'M')          /* Check for errors. */
      {
        printf("Unknown designator : %02X\n", (unsigned char) *p);
        break;
      }

    seg = seg + MCB_size(p) + 1;  /* Move to next MCB */
  }


  printf("\n");  /* Finish table. */
}


unsigned find_start()
{
  asm push es;
  asm mov  ah, 0x52;
  asm int  0x21;
  asm sub  bx, 2;
  asm mov  ax, es:[bx];
  asm pop  es;

  return _AX;
}


/*
 * And finally, the main() procedure.
 *
 */
main()
{
  start = find_start();      /* Find a start MCB. */

  trace_chain(start);

  exit(0);                   /* A clean exit. */
}
#endif

